/*## (c) SECURITY EXPLORATIONS    2011 poland                                #*/
/*##     http://www.security-explorations.com                                #*/

/* RESEARCH MATERIAL:	SRP-2018-01                                           */
/* SlimCORE disassembler                                                      */

/* THIS SOFTWARE IS PROTECTED BY DOMESTIC AND INTERNATIONAL COPYRIGHT LAWS    */
/* UNAUTHORISED COPYING OF THIS SOFTWARE IN EITHER SOURCE OR BINARY FORM IS   */
/* EXPRESSLY FORBIDDEN. ANY USE, INCLUDING THE REPRODUCTION, MODIFICATION,    */
/* DISTRIBUTION, TRANSMISSION, RE-PUBLICATION, STORAGE OR DISPLAY OF ANY      */
/* PART OF THE SOFTWARE, FOR COMMERCIAL OR ANY OTHER PURPOSES REQUIRES A      */
/* VALID LICENSE FROM THE COPYRIGHT HOLDER.                                   */

/* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS    */
/* OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,*/
/* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL    */
/* SECURITY EXPLORATIONS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, */
/* WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF  */
/* OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE     */
/* SOFTWARE.                                                                  */

import java.lang.*;
import java.util.*;

public class Opcode {
 static String[] map[]={
  //00 opcodes
  {"mov",  "0000 reg1 0000 reg2 0011 1100", "reg1 , reg2"},
  {"swap", "0000 reg1 reg2 0000 1100 0000", "reg1 , reg2"},

  //01 opcodes
  {"shl",  "0001 reg1 reg2 0000 000 imm5",  "reg1 , reg2 , # imm5"},
  {"shr",  "0001 reg1 reg2 0000 001 imm5",  "reg1 , reg2 , # imm5"},

  //02 opcodes
  {"add",  "0010 reg1 reg2 reg3 imm8", "reg1 , reg2 , reg3 , # imm8"},

  //03 opcodes
  {"sub",  "0011 reg1 reg2 reg3 imm8", "reg1 , reg2 , reg3 , # imm8"},

  //04 opcodes
  {"and", "0100 reg1 reg2 reg3 00000000", "reg1 , reg2 , reg3"},
  {"and", "0100 reg1 reg2 0000 imm8",     "reg1 , reg2 , # imm8"},
  {"tst", "0100 0000 reg  0000 imm8",     "reg , # imm8"},
  {"tst", "0100 0000 reg1 reg2 00000000", "reg1 , reg2"},

  //05 opcodes
  {"or",  "0101 reg1 reg2 reg3 00000000", "reg1 , reg2 , reg3"},
  {"or",  "0101 reg1 reg2 0000 imm8",     "reg1 , reg2 , # imm8"},
  {"tst", "0101 0000 0000 reg 00000000",  "reg , reg"},

  //06 opcodes
  {"xor", "0110 reg1 reg2 reg3 00000000", "reg1 , reg2 , reg3"},
  {"xor", "0110 reg1 reg2 0000 imm8",     "reg1 , reg2 , # imm8"},

  //07 opcodes
  {"and",    "0111 reg1 reg2 11 bitnum 00000", "reg1 , reg2 , bitnum"},
  {"mov",    "0111 reg1 reg2 11 bitnum shift", "reg1 , ( reg2 >> shift ) & bitnum"},
  {"mov",    "0111 reg1 reg2 00 bitnum shift", "reg1 , ( reg2 & bitnum ) << shift"},
  {"mov",    "0111 reg1 reg2 01 bitnum 00000", "reg1 , reg2 & bitnum"},
  {"mov",    "0111 reg1 reg2 00 bitnum 00000", "reg1 , reg2 & bitnum"},
  {"mov",    "0111 reg1 reg2 01 00001  shift", "reg1 , reg2 &0x01 << shift"},
  {"movzx",  "0111 reg1 reg2 01 01000 00000",  "reg1 , reg2 &0xff"},
  {"movhi",  "0111 reg1 reg2 00 bitnum 10000", "reg1 , ( reg2 & bitnum ) <<16"},
  {"movhi",  "0111 reg1 reg2 00 10000 10000",  "reg1 , reg2 <<16"},
  {"bitset", "0111 reg1 reg2 00 00001 shift",  "reg1 , reg2 &0x01 << shift"},
  {"bitclr", "0111 reg1 0000 00 00001 shift",  "reg1 , #1 << shift"},
  {"bitval", "0111 reg1 reg2 11 00001 shift",  "reg1 , reg2 , #1 << shift"},
  {"bittst", "0111 0000 reg 11 00001 shift",   "reg , #1 << shift"},

  //08 opcodes
  {"jmp",   "1000 0100 0000 reg 00000000", "reg"},
  {"j",     "1000 1100 0000 target",       "target"},
  {"jz",    "1000 1000 0001 target",       "target"},
  {"jne",   "1000 1100 0001 target",       "target"},
  {"js",    "1000 1000 0010 target",       "target"},
  {"jns",   "1000 1100 0010 target",       "target"},
  {"jxx1",  "1000 1111 0010 target",       "target"},
  {"wait1", "1000 1110 0001 target",       ""},
  {"wait2", "1000 1101 1000 target",       ""},

  //09 opcodes
  {"je",    "1001 1000 0001 target", "target"},
  {"jb",    "1001 1000 0100 target", "target"},
  {"jae",   "1001 1100 0100 target", "target"},
  {"jbe",   "1001 1000 0111 target", "target"},
  {"jne",   "1001 1100 0001 target", "target"},
  {"jns",   "1001 1100 0010 target", "target"},
  {"js",    "1001 1000 0010 target", "target"},
  {"jzs",   "1001 1000 0011 target", "target"},
  {"jxx2",  "1001 1011 0100 target", "target"},
  {"jxx3",  "1001 1010 1000 target", "target"},
  {"wait3", "1001 1110 0001 target", ""},
  {"wait4", "1001 1101 1000 target", ""},

  //0a opcodes
  {"ld", "1010 reg1 0000 reg2 mimm8", "reg1 , [ reg2 + mimm8 ]"},
  {"ld", "1010 reg1 0001 reg2 imm8",  "reg1 , [ reg2 ] , reg2 +=# imm8"},

  //0b opcodes
  {"st", "1011 0000 reg1 reg2 mimm8", "reg1 , [ reg2 + mimm8 ]"},
  {"st", "1011 0001 reg1 reg2 imm8",  "reg1 , [ reg2 ] , reg2 +=# imm8"},

  //0c opcodes
  {"cmp", "1100 0000 reg 0000 imm8", "reg , # imm8"},

  //0d opcodes
  {"jmp",             "1101 0000 target_hi 0001 target_lo", "target"},
  {"bitsrch topmost", "1101 reg1 reg2 0000 0100 0000",      "reg1 , reg2"},
  {"sync1",           "1101 0000 0000 0000 1000 0000",      ""},
  {"sync2",           "1101 0000 0000 0000 1001 0000",      ""},
  {"rpt",             "1101 0000 0000 0000 0000 n",         "n"},
  {"rpt 16",          "1101 0000 0001 0000 0000 0000",      ""},

  //0e opcodes
  {"mov", "1110 reg imm16", "reg , # imm16"},

  //0f opcodes
  {"copAES",      "1111 0101 0100 0000 0000 0000", ""},
  {"copTDES",     "1111 1010 0100 0000 0000 0000", ""},
  {"copAES_dma",  "1111 0100 0100 0000 0000 0000", ""},
  {"copTDES_dma", "1111 1000 0100 0000 0000 0000", ""}
 };

 static {
  components=new Hashtable();
  add_component("reg",4);
  add_component("reg1",4);
  add_component("reg2",4);
  add_component("reg3",4);
  add_component("imm5",5);
  add_component("imm8",8);
  add_component("mimm8",8);
  add_component("imm16",16);
  add_component("shift",5);
  add_component("bitnum",5);
  add_component("target",12);
  add_component("target_hi",8);
  add_component("target_lo",4);
  add_component("n",4);

  build_decoders();
  build_helper();
 }

 static Hashtable components;

 static Decoder decoders[];
 static DecoderItem helper[];

 static Decoder UNK_DECODER=new Decoder("","","");

 //parameters for extraction of a given opcode component value
 //comp_value=(opcode>>shift)&bitmask
 public static class ComponentAccessor {
  int shift;
  int bitmask;

  public int get(int opcode) {
   return (opcode>>shift)&bitmask;
  }

  public ComponentAccessor(int shift,int size) {
   this.shift=shift;
   this.bitmask=Decoder.mask(size-1);
  }
 }

 public static class Decoder {
  String name;
  String[] format;
  String[] print_tokens;
  int weight;
  int val;
  int mask;
 
  static int masks[];

  //component accessors for this opcode format
  //a map of component name -> ComponentAccessor instance
  Hashtable accessors;

  static {
   masks=new int[32];
   
   for(int i=0;i<masks.length;i++) {
    if (i==0) masks[i]=1;
     else masks[i]=masks[i-1]|(1<<i);
   }
  }

  public String[] str2tokens(String s) {
   s=s.trim();

   StringTokenizer st=new StringTokenizer(s," ", false);
   int cnt=st.countTokens();

   String tokens[]=new String[cnt];

   for(int i=0;i<cnt;i++) {
    tokens[i]=st.nextToken();
   }            

   return tokens;
  }

  public int weight(String[] format) {
   int weight=0;

   for(int i=0;i<format.length;i++) {
    String component=format[i];

    if (component.startsWith("0")||component.startsWith("1")) {
     weight+=component.length();
    }
   }

   return weight;
  }

  public static int mask(int bitnum) {
   return masks[bitnum];
  }

  public int mask(int bitnum,int len) {
   int mask=masks[bitnum];

   int nbitnum=bitnum-len-1;

   if (nbitnum>=0) {
    mask^=masks[nbitnum];
   }

   return mask;
  }

  public void calc_value_and_mask() {
   mask=0;
   val=0;

   int shift=24;

   for(int i=0;i<format.length;i++) {
    String component=format[i];

    int size=component_size(component);

    if (size>0) {
     shift-=size;
     val<<=size;
    } else {
     mask|=mask(shift-1,component.length()-1);
     shift-=component.length();
     val<<=component.length();
     val|=Integer.parseInt(component,2);
    }
   }   
  }

  public int opclass() {
   if (format.length==0) return -1;

   int opclass=Integer.parseInt(format[0],2);

   return opclass;
  }

  public boolean match(int opcode) {
   if ((opcode&mask)==(val&mask)) return true;
    else return false;
  }

  public boolean is_jump() {
   if ((print_tokens.length==1)&&(print_tokens[0].equals("target"))) return true;
    else return false;
  }

  public boolean is_load_or_store() {
   if (name.equals("ld")||name.equals("st")) return true;
    else return false;
  }

  public int jump_target(int opcode) {
   if (is_jump()) {
    if (name.equals("jmp")) {
     ComponentAccessor accessor1=(ComponentAccessor)accessors.get("target_hi");
     ComponentAccessor accessor2=(ComponentAccessor)accessors.get("target_lo");

     int target=(accessor1.get(opcode)<<4)|accessor2.get(opcode);
     return target;
    } else {
     ComponentAccessor accessor=(ComponentAccessor)accessors.get("target");

     int target=accessor.get(opcode);
     return target;
    }
   } else return -1;
  }

  public int get_component(int opcode,String component) {
   ComponentAccessor accessor=(ComponentAccessor)accessors.get(component);

   if (accessor!=null) {
    int val=accessor.get(opcode);
    return val;
   }

   return -1;
  }

  public String print_component(String component,int val) {
   String print="";

   if (component.startsWith("reg")) {
    print+="r"+val;
   } else

   if (component.equals("imm5")) {
    print+=hex_value(val,2);
   } else

   if (component.equals("imm8")) {
    print+=hex_value(val,2);
   } else

   if (component.equals("mimm8")) {
    print+=hex_value(val,4);
   } else

   if (component.equals("imm16")) {
    print+=hex_value(val,4);
   } else

   if (component.equals("shift")) {
    print+=hex_value(val,2);
   } else

   if (component.equals("bitnum")) {
    print+="0x"+hex_value(mask(val-1),4);
   } else

   if (component.equals("target")) {
    print+="l_"+hex_value(val,4);
   } else

   if (component.equals("n")) {
    print+=hex_value(val,1);
   }

   return print;
  }

  public String print(int opcode) {
   String print="";

   print+=name;

   if (print_tokens.length>0) print+=" ";

   for(int i=0;i<print_tokens.length;i++) {
    String token=print_tokens[i];

    ComponentAccessor accessor=(ComponentAccessor)accessors.get(token);

    if (accessor!=null) {
     print+=print_component(token,accessor.get(opcode));
    } else {
     if (name.equals("jmp")) {
      int val=jump_target(opcode);
      print+=print_component(token,val);
     } else {
      print+=token;
     }
    }
   }

   return print;
  }

  public void setup_accessors() {
   accessors=new Hashtable();

   int shift=24;

   for(int i=0;i<format.length;i++) {
    String component=format[i];

    int size=component_size(component);

    if (size>0) {
     shift-=size;
//     System.out.println(" - "+component+" shift "+Integer.toHexString(shift)+" mask "+Integer.toHexString(mask(size-1)));
     accessors.put(component,new ComponentAccessor(shift,size));
    } else {
     shift-=component.length();
    }
   }   
  }

  public Decoder(String name,String sformat,String sprint_tokens) {
   this.name=name;
   this.format=str2tokens(sformat);
   this.print_tokens=str2tokens(sprint_tokens);

   weight=weight(format);

   calc_value_and_mask();
//   System.out.println(opclass()+" "+name+" val "+Integer.toHexString(val)+" mask "+Integer.toHexString(mask));

   setup_accessors();
  }

  public Decoder(String[] opcode_desc) {
   this(opcode_desc[0],opcode_desc[1],opcode_desc[2]);
  }
 }

 public static class DecoderItem {
  Decoder decoder;
  DecoderItem next;

  public DecoderItem(Decoder decoder) {
   this.decoder=decoder;
  }

  public DecoderItem(Decoder decoder,DecoderItem next) {
   this.decoder=decoder;
   this.next=next;
  }

  public void add(Decoder d) {
   if (decoder.weight>d.weight) {
    if (next!=null) {
     next.add(d);
    } else {
     next=new DecoderItem(d);
    }
   } else {
    next=new DecoderItem(decoder,next);
    decoder=d;
   }
  }
 }

 static String hex_value(int val,int max) {
  String s="";
  for(int i=0;i<max;i++) {
   s+="0";
  }

  s+=Integer.toHexString(val);

  return s.substring(s.length()-max);
 }

 static void add_component(String name,int size) {
  components.put(name,new Integer(size));
 }

 static int component_size(String name) {
  Integer ival=(Integer)components.get(name);

  if (ival!=null) return ival.intValue();
   else return 0;
 }

 static void build_decoders() {
  int size=map.length;

  decoders=new Decoder[size];

  for(int i=0;i<size;i++) {
   decoders[i]=new Decoder(map[i]);
  }
 }

 static void build_helper() {
  helper=new DecoderItem[16];

  for(int i=0;i<decoders.length;i++) {
   Decoder decoder=decoders[i];
   int opclass=decoder.opclass();

   if (helper[opclass]==null) {
    helper[opclass]=new DecoderItem(decoder);
   } else {
    helper[opclass].add(decoder);
   }
  }
 }

 public static Decoder parse(int opcode) {
  int opclass=(opcode>>20)&0x0f;

  DecoderItem item=helper[opclass];

  while(item!=null) {
   Decoder d=item.decoder;

   if (d.match(opcode)) return d;
    else item=item.next;
  }

  return UNK_DECODER;
 }

 public static String ins_desc(int opcode) {
  Opcode.Decoder d=Opcode.parse(opcode);
  return d.print(opcode);
 }
 
}
